# -*- coding: UTF-8 -*-
"""
Caches for memoization
"""

from __future__ import absolute_import, unicode_literals


__all__ = ("LRUCache", "FIFOCache", "TimedCache")


from collections import OrderedDict
from time import time


class LRUCache(OrderedDict):
    """Self-pruning cache using an LRU strategy.

    If instantiated with a ``max_size`` other than ``0``, will
    automatically prune the least-recently-used (LRU) key/value
    pair when inserting an item after reaching the specified size.

    An item is considered to be "used" when it is inserted or
    accessed, at which point its position in recently used
    queue is updated to the most recent.

    Supports all standard dictionary methods.

    :param int max_size: maximum number of entries to save
        before pruning
    """

    def __init__(self, max_size=0, *args, **kwargs):
        super(LRUCache, self).__init__(*args, **kwargs)
        self._max_size = max_size

    def __getitem__(self, key, **kwargs):
        value = OrderedDict.__getitem__(self, key)
        del self[key]
        OrderedDict.__setitem__(self, key, value, **kwargs)  # type: ignore
        return value

    def __setitem__(self, key, value, **kwargs):
        if key in self:
            del self[key]
        OrderedDict.__setitem__(self, key, value, **kwargs)  # type: ignore
        if self._max_size and len(self) > self._max_size:
            self.popitem(last=False)


class FIFOCache(OrderedDict):
    """Self-pruning cache using a FIFO strategy

    If instantiated with a ``max_size`` other than ``0``, will
    automatically prune the least-recently-inserted key/value
    pair when inserting an item after reaching the specified
    size.

    Supports all standard dictionary methods.

    :param int max_size: maximum number of entries to save
        before pruning
    """

    def __init__(self, max_size=0, *args, **kwargs):
        super(FIFOCache, self).__init__(*args, **kwargs)
        self._max_size = max_size

    def __setitem__(self, key, value, **kwargs):
        OrderedDict.__setitem__(self, key, value)
        if self._max_size and len(self) > self._max_size:
            self.popitem(last=False)


class TimedCache(dict):
    """Self-pruning cache whose entries can be set to expire

    If instantiated with a ``max_age`` other than ``0``, will
    consider entries older than the specified age to be invalid,
    removing them from the cache upon an attempt to access them
    and returning as though they do not exist.

    Supports all standard dictionary methods.

    :param int max_age: age in seconds beyond which entries
        should be considered invalid. The default is 0, which
        means that entries should be stored forever.
    """

    def __init__(self, max_age=0, *args, **kwargs):
        super(TimedCache, self).__init__(*args, **kwargs)
        self._max_age = max_age

    def __getitem__(self, key):
        value, last_time = dict.__getitem__(self, key)
        now = time()
        if self._max_age and now - last_time > self._max_age:
            del self[key]
            raise KeyError(key)
        else:
            return value

    def __setitem__(self, key, value):
        now = time()
        dict.__setitem__(self, key, (value, now))

    def __contains__(self, key):
        try:
            self.__getitem__(key)
        except KeyError:
            return False
        return True



# -*- coding: UTF-8 -*-
"""
version.py module

The version set here will be automatically incorporated into setup.py
and also set as the __version__ attribute for the package.
"""

from __future__ import absolute_import, unicode_literals

__version_info__ = (2, 0, 1)
__version__ = ".".join([str(ver) for ver in __version_info__])



# -*- coding: UTF-8 -*-
"""
Public interface for decorators sub-package
"""

__all__ = (
    "after",
    "before",
    "construct_decorator",
    "decorate",
    "export",
    "instead",
    "Decorated",
    "intercept",
    "log_call",
    "memoize",
)

from .generic import (
    after,
    before,
    construct_decorator,
    decorate,
    Decorated,
    instead,
)
from .ready_to_wear import (
    export,
    intercept,
    log_call,
    memoize,
)



# -*- coding: UTF-8 -*-
"""All decorators are exposed directly here.

In addition, decorators may be imported from `pydecor.decorators`.

Non-decorator utilities like caches and functions are contained
in their own modules and not exposed here.
"""

from ._version import __version__, __version_info__  # noqa
from . import decorators
from .decorators import *  # noqa

__all__ = decorators.__all__



# -*- coding: UTF-8 -*-
"""
Private interface utilities for PyDecor
"""

from __future__ import absolute_import, unicode_literals


__all__ = ("ClassWrapper",)


from functools import partial
from inspect import isclass
from logging import getLogger
from sys import version_info

from inspect import isfunction, ismethod


log = getLogger(__name__)


PY2 = version_info < (3, 0)


def get_fn_args(decorated, args):
    """Strip "self" and "cls" variables from args.

    For now, this avoids assuming that the "self" variable is called
    "self" on instance methods, although if this winds up being
    problematic for edge use cases, I might move to that model
    and make it clear that instance methods for which "self"
    isn't called "self" won't have their "self" argument stripped.
    """
    fn_args = args
    decor_name = (
        decorated.__name__ if hasattr(decorated, "__name__") else str(decorated)
    )

    # Check if the wrapped function is an attribute on args[0]
    if args and decor_name in dir(args[0]):

        # Check if args[0] is a class reference
        if isclass(args[0]):
            # Check if the wrapped function is a classmethod on the
            # class' class dict
            if type(args[0].__dict__[decor_name]) is classmethod:
                # The first argument is a reference to the class
                # containing the decorated function
                fn_args = args[1:]

        # Check if args[0] is a function defined on the class
        else:
            cls_dict = args[0].__class__.__dict__

            # Check if the wrapped function is a function on the class'
            # class dict (instance methods are functions on the class
            # dict, while classmethods and staticmethods are their own
            # types)
            if decor_name in cls_dict and isfunction(cls_dict[decor_name]):
                # The first argument is probably a "self" variable
                fn_args = args

    return fn_args


class ClassWrapper(object):
    """A generic class wrapper for decorating class functions/methods

    To be used in the context of a decorator that should percolate
    down to methods if used at the class level.

    Intended usage is to create a **new** class, for example by
    using the ``type()`` function, in which the ``wrapped``
    class attribute is replaced by a reference to the class
    being wrapped

    Usage:

    .. code:: python


        def meth_decorator(meth):
            '''The decorator that should be applied to class methods'''

            def wrapper(*args, **kwargs):
                print('performing wrapper tasks')
                func(*args, **kwargs)
                print('finished wrapper tasks')

            return wrapper


        def class_decorator(cls):
            '''Applies meth_decorator to class methods/functions'''
            wrapper = ClassWrapper.wrap(cls, meth_decorator)
            return wrapper


        @class_decorator
        class DecoratedClass:

            def automatically_decorated_method(self):
                pass

    """

    __wrapped__ = None
    __decorator__ = None
    __decoropts__ = None

    def __init__(self, *args, **kwargs):
        self.__wrapped__ = self.__wrapped__(*args, **kwargs)  # type: ignore

    def __getattribute__(self, item):

        if item in ("__wrapped__", "__decorator__", "__decoropts__"):
            return object.__getattribute__(self, item)

        wrapped = object.__getattribute__(self, "__wrapped__")
        attr = getattr(wrapped, item)

        if attr is None:
            raise AttributeError

        if ismethod(attr) or isfunction(attr):
            cls = object.__getattribute__(self, "__class__")
            decoropts = object.__getattribute__(cls, "__decoropts__")
            decor = object.__getattribute__(cls, "__decorator__")

            if decoropts["instance_methods_only"]:
                cls_attr = object.__getattribute__(cls, item)
                if type(cls_attr) is classmethod:
                    return attr
                if type(cls_attr) is staticmethod:
                    return attr

            return decor(attr)

        return attr

    @classmethod
    def _get_class_attrs(cls, wrapped, decorator, instance_methods_only):
        """Get the attrs for a new class instance

        :param type wrapped: a reference to the class to be wrapped
        :param Union[FunctionType,MethodType,type] decorator:
            the decorator to apply to the
            functions and methods of the wrapped class
        """
        attrs = {}

        for k, v in wrapped.__dict__.items():
            if not k.startswith("__"):
                if not instance_methods_only:
                    if type(v) is classmethod:
                        attrs[k] = partial(decorator(v.__func__), wrapped)
                    elif type(v) is staticmethod:
                        if PY2:
                            attrs[k] = staticmethod(
                                partial(decorator(v.__func__))
                            )
                        else:
                            attrs[k] = decorator(v.__func__)
                    else:
                        attrs[k] = v
                else:
                    attrs[k] = v

        attrs.update(
            {
                "__wrapped__": wrapped,
                "__decorator__": decorator,
                "__decoropts__": {  # type: ignore
                    "instance_methods_only": instance_methods_only
                },
            }
        )

        return attrs

    @classmethod
    def wrap(cls, wrapped, decorator, instance_methods_only=False):
        """Return a new class wrapping the passed class

        :param type wrapped: a reference to the class to be wrapped
        :param Union[FunctionType,MethodType,type] decorator:
            the decorator to apply to the
            functions and methods of the wrapped class
        """
        name = "Wrapped{}".format(wrapped.__name__)
        if PY2:
            name = str(name)

        return type(
            name,
            (cls,),
            # {'__decorator__': decorator}
            cls._get_class_attrs(wrapped, decorator, instance_methods_only),
        )



# EMACS settings: -*- coding: utf-8 -*-
"""
Decorators controlling visibility of entities in a Python module.
"""

__all__ = ("export",)
__api__ = __all__

import sys
import typing as t


T = t.TypeVar("T", bound=t.Callable)


def export(entity: T) -> T:
    """Register the given function or class as publicly accessible in a module.

    Creates or updates the `__all__` attribute in the module in which the
    decorated entity is defined to include the name of the decorated
    entity.

    Example:

    `to_export.py`:

    .. code:: python

        from pydecor import export

        @export
        def exported():
            pass

        def not_exported():
            pass


    `another_file.py`

    .. code:: python

        from .to_export import *

        assert "exported" in globals()
        assert "not_exported" not in globals()


    :param Union[Type, types.FunctionType] entity: the function or class
        to include in `__all__`
    """
    # * Based on an idea by Duncan Booth:
    #   http://groups.google.com/group/comp.lang.python/msg/11cbb03e09611b8a
    # * Improved via a suggestion by Dave Angel:
    #   http://groups.google.com/group/comp.lang.python/msg/3d400fb22d8a42e1

    if not hasattr(entity, "__module__"):
        raise TypeError(
            (
                "{entity} has no __module__ attribute. Please ensure it is "
                "a top-level function or class reference defined in a module."
            ).format(entity=entity)
        )

    if hasattr(entity, "__qualname__"):
        if any(i in entity.__qualname__ for i in (".", "<locals>", "<lambda>")):
            raise TypeError(
                "Only named top-level functions and classes may be exported, "
                "not {}".format(entity)
            )

    if not hasattr(entity, "__name__") or entity.__name__ == "<lambda>":
        raise TypeError(
            "Entity must be a named top-level funcion or class, not {}".format(
                type(entity)
            )
        )

    try:
        module = sys.modules[entity.__module__]
    except KeyError:
        raise ValueError(
            (
                "Module {} is not present in sys.modules. Please ensure "
                "it is in the import path before calling export()."
            ).format(entity.__module__)
        )
    if hasattr(module, "__all__"):
        if entity.__name__ not in module.__all__:  # type: ignore
            module.__all__ = module.__all__.__class__(  # type: ignore
                (*module.__all__, entity.__name__)  # type: ignore
            )
    else:
        module.__all__ = [entity.__name__]  # type: ignore

    return entity
